/*  linemon - a simple TCP/IP connectivity test/stats program
    to monitor the connectivity between two hosts.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Copyright 2001 by Jarle (jgaa) Aase, jgaa@jgaa.com
    http://www.jgaa.com
*/

#include "StdAfx.h"


#define PROGRAM "LineMonitor"
#define VERSION "1.01"

#ifdef WIN32

#   include "MyNtService.h"

#else

#   include <stdio.h>
#   include <sys/types.h>
#   include <sys/stat.h>
#   include <unistd.h>
#   include <fcntl.h>
#   include <sys/socket.h>
#   include <netinet/in.h>
#   include <sys/time.h>
#   include <netinet/tcp.h>
#   include <arpa/inet.h>
#   include <netdb.h>
#   include <errno.h>
#   include <stdlib.h>
#   include <signal.h>

#endif

#include <iostream>
#include <fstream>
#include <vector>
#include <getopt.h>
#include <time.h>

using namespace std;

int my_port = 1030;
long my_host;
int verbose_flag = 0;
int debug_flag = 0;
int server_mode = 0;
int daemon_mode = 0;
int poll_delay = 1; // delay in seconds
int loop_delay = 1;
int alarm_delay = 10;
char *myname = "";
char *hostname = NULL;
const char *logfile_name = NULL;
bool do_continue = true;

#ifdef WIN32
int reply_delay = 0;
const char *service_name = "LineMonitor";
#endif

#ifdef WIN32
/* Microsoft is working very hard to push all Wintendo developers 
   over to VisualBasic. One of the methods is to make the
   "standard" API's worthless, so that developers have to write
   as much as possible from scratch if they choose to use C++ or C.
   std::ofstream is implemented without any way to turn on file-sharing,
   so that the stream classes are useless for any logging functions.
   We have to write our own class...
*/
#   include "ofstream_shared.h"

    ofstream_shared log_file;
#   define ENDL "\r\n"

#else
    ofstream log_file;
#   define ENDL endl
#endif

struct sockaddr socaddr;
struct sockaddr_in *pin = (struct sockaddr_in *)&socaddr;
int sockaddr_len = sizeof(sockaddr);

#ifdef WIN32
#   define ISSLASH(a) (a == '\\')
#   define GET_SOCKET_ERROR WSAGetLastError()
#   define EWOULDBLOCK  WSAEWOULDBLOCK
typedef int socklen_t;
#else
#   define ISSLASH(a) (a == '/')
#   define GET_SOCKET_ERROR errno
#   define SOCKET int
#   define BOOL int
#   define closesocket(a) close(a)
#endif

void usage();
int do_client();
int do_server();
char *get_time();
int do_process();
void show_gpl();
SOCKET create_socket(bool doBind);


void usage()
{
    cout << myname << " [options] hostname" << endl;
    cout << "Options: " << endl;
    cout << " -p  --port=#        Set port number. Default is 1030" << endl;
    cout << " -v  --version       Show version number and quit" << endl;
    cout << "     --verbose       Turn on verbose mode" << endl;
    cout << "     --debug         Turn on debug messages" << endl;
    cout << "     --server        Server mode (default is client mode)" << endl;
    cout << "     --daemon        Runs as a daemon" << endl;
    cout << " -l  --log-file=name Name of logfile" << endl;
    cout << " -h  --help          Show this help" << endl;
    cout << " -w  --wait=delay    Delay between probes in seconds" << endl;

#ifdef WIN32
    cout << " -r  --reply-delay=# Delay before pong reply (server mode)" << endl;
    cout << " -s  --service-name=name Sets the service name. Used with --install" << endl;
    cout << " -i  --install       Install as a NT service (and quit)" << endl;
    cout << " -u  --uninstall     Uninstall as a NT service (and quit)" << endl;
#endif

    cout << " -t  --timeout=sec   Timeout delay in seconds" << endl;
    cout << " -L  --license       Show licensing terms (GPL)" << endl;
    cout << endl;
    cout << "hostname is optional in server-mode" << endl;
    
    exit(-1);
}

#ifdef WIN32

BOOL WINAPI win32_break(DWORD dwCtrlType)
{
    cerr << "**** Interrupted. Shutting down" << endl;
    
    do_continue = false;

    return true;
}

#else
void sig_alarm(int sig)
{
    ;
}

void sig_break(int signal)
{
    cerr << "**** Interrupted. Shutting down" << endl;
    
    do_continue = false;
}

#endif



int main(int argc, char* argv[])
{
#ifdef WIN32
    bool do_uninstall = false;
    bool do_install = false;
#endif
    char *p = argv[0];
    if (p)
    {
        while (*p)
        {
            if (ISSLASH(*p))
                myname = p + 1;
            ++p;
        }
    }
    
    if (!*myname)
        myname = argv[0];
    
    struct option options[] = 
    {
        {"verbose", no_argument, &verbose_flag, 1},
        {"version", no_argument, NULL, 'v'},
        {"install", no_argument, NULL, 'i'},
        {"uninstall", no_argument, NULL, 'u'},
        {"port", required_argument, NULL, 'p'},
        {"server", no_argument, &server_mode, 1},
        {"daemon", no_argument, &daemon_mode, 1},
        {"debug", no_argument, &debug_flag, 1},
        {"log-file", required_argument, NULL, 'l'},
        {"wait", required_argument, NULL, 'w'},
        {"timeout", required_argument, NULL, 't'},
        {"license", no_argument, NULL, 'L'},

#ifdef WIN32
        {"reply-delay", required_argument, NULL, 'r'},
        {"service-name", required_argument, NULL, 's'},
#endif

        {"help", no_argument, NULL, 'h'},
    };
    
    int ch = 0;
    while((ch = getopt_long(argc, argv, "Lhvp::w::t::l::r::t::s::", options, NULL)) >= 0)
    {
        switch(ch)
        {
        case 'v':
            cout << PROGRAM << " " << VERSION << endl;
            return 0;
            
        case 'w':
            loop_delay = atoi(optarg);
            if (verbose_flag)
                cout << "Setting wait delay to " << loop_delay << " seconds" << endl;
            break;
            
        case 'p':
            my_port = atoi(optarg);
            if (verbose_flag)
                cout << "Setting port to " << my_port << endl;
            break;
            
        case 't':
            poll_delay = atoi(optarg);
            break;

#ifdef WIN32
        case 'r':
            reply_delay = atoi(optarg);
            break;

        case 'i':
            do_install = true;
            break;

        case 'u':
            do_uninstall = true;
            break;
        case 's':
            service_name = optarg;
            break;
#endif

        case 'L':
            show_gpl();
            return 0;
            
        case 'l':
            if (verbose_flag)
                cout << "Opening logfile " << optarg << endl;
            
            logfile_name = optarg;
            log_file.open(optarg, ios::out|ios::app);

            if (!log_file.is_open())
            {
                cerr << "Failed to open logfile " << optarg << endl;
                return -1;
            }
            break;
            
        case 0:
            break;
            
        default:
            usage();
        }
    }
    
#ifdef WIN32
    WSADATA wsaData;
    if (WSAStartup(MAKEWORD(2, 0), &wsaData ) == -1)
    {
        cerr << "Failed to initialize winsock. " << GET_SOCKET_ERROR << endl;
        return -1;
    }

    SetConsoleCtrlHandler(win32_break, true);

#else
    
    if (daemon_mode)
    {
	if (verbose_flag)
	    cout << "Switching to daemon mode" << endl;

        switch(fork())
        {
        case -1:
            cerr << "Failed to fork " << errno << endl;
            return -1;
        case 0: // child
            close(STDIN_FILENO);
            close(STDOUT_FILENO);
            close(STDERR_FILENO);
            if (setsid() == -1)
                return -1;
            if (fork() != 0)
                return 0;
            chdir("/");
            umask(0);
            freopen("/dev/null", "r", stdin);
            freopen("/dev/null", "w", stdout);
            freopen("/dev/console", "w", stderr);
            break;
        default:
            return 0;
        }
        
    }
    
    signal(SIGALRM, sig_alarm);
    signal(SIGQUIT, sig_break);
    signal(SIGINT, sig_break);
#endif // WIN32
    
    
    if (argc != (optind +1))
    {
        if (!server_mode)
        {
            cerr << "Missing hostname argument" << endl;
            return -1;
        }
    }
    else
    {
        hostname = argv[argc -1];
        my_host = ntohl(inet_addr(hostname));
        if (((SOCKET)INADDR_NONE) == my_host)
        {
            if (verbose_flag)
            {
                cout << "Looking up hostname " << hostname << " ... ";
                cout.flush();
            }
            
            struct hostent *pent = gethostbyname(hostname);
            
            if (NULL == pent)
            {
                if (verbose_flag)
                    cout << "FAILED" << endl;
                
                cerr << optarg << ": Unknown my_host" << endl;
                return -1;
            }
            my_host = *(long *)pent->h_addr;
            
            if (verbose_flag)
            {
                in_addr my_addr;
                my_addr.s_addr = my_host;
                cout << inet_ntoa(my_addr) << endl;
            }
        }
    }

#ifdef WIN32
    if (daemon_mode)
    {
        MyNtService my_service;
        if (!my_service.StartService())
        {
            cerr << "Cannot start as NT/Windows 2000 service!" << endl;
            exit(-1);
        }
    }

    if (do_uninstall)
    {
        MyNtService my_service;
        if (!my_service.IsInstalled())
        {
            cerr << "The application is not installed as a NT service!" << endl;
            return -1;
        }
        if (my_service.Uninstall())
        {
            cout << "Successfully uninstalled as NT service" << endl;
            if (!do_install)
                return 0;
        }
        else
        {
            cerr << "Failed to uninstall as NT servive!" << endl;
            return -1;
        }
    }

    if (do_install)
    {
        string myargs;
        MyNtService my_service;
        if (my_service.IsInstalled())
        {
            cerr << "The application is already installed." << endl;
            return -1;
        }
        
        char buffer[32];
        
        myargs += "--daemon ";
        myargs += "--port=";
        myargs += itoa(my_port, buffer, 10);
        myargs += " --wait=";
        myargs += itoa(loop_delay, buffer, 10);
        myargs += " ";
        
        if (server_mode)
            myargs += "--server ";
        
        if (logfile_name)
        {
            myargs += "--log-file=";
            myargs += logfile_name;
            myargs += " ";
        }
        
        if (hostname)
            myargs += hostname;
        
        if (my_service.Install(myargs.c_str()))
        {
            cout << "Successfully installed as a NT service." 
                << endl
                << "You must now enter the service applet in the "
                "control panel and configure the startup mode."
                << endl;
            return 0;
        }
        {
            cerr << "Failed to install as NT servive!" << endl;
            return -1;
        }
    }
#endif // WIN32
    
    return do_process();
}

int do_process()
{
    cout << PROGRAM 
        << " version " 
        << VERSION
        << ", Copyright (C) 2001 Jarle (jgaa) Aase" 
        << endl
        << PROGRAM 
        << " comes with ABSOLUTELY NO WARRANTY"
        << endl
        << "This is free software, and you are welcome to redistribute it "
        << endl
        << "under certain conditions;"
        << endl
        << myname << " --license for details."
        << endl;

    if (server_mode)
        return do_server();
    
    return do_client();
}

char *get_time()
{
    static char buf[64];
    time_t now = time(NULL);
    struct tm *ptm = localtime(&now);
    if (!ptm)
        return "***notime***";
    
    strftime(buf, sizeof(buf), "%Y-%m-%d %H:%M", ptm);
    
    return buf;
}

int set_async_mode(SOCKET sck, bool makeAsync)
{    
    
#if WIN32
    unsigned long sck_mode = makeAsync;
    if (::ioctlsocket(sck, FIONBIO, &sck_mode))
        return -1;
#else
    int flags = 0;
    
    if ((flags = ::fcntl(sck, F_GETFL, 0)) != 0)
        return -1;
    
    if (makeAsync)
        flags |= O_NONBLOCK;
    else
        flags &= ~O_NONBLOCK;
    
    if (fcntl(sck, F_SETFL, flags) != 0)
        return -1;
#endif

    return 0;
}

void sleep(int seconds)
{
#ifdef WIN32
    Sleep(seconds * 1000);
#else
    alarm(seconds);
    pause();
#endif
}

SOCKET create_socket(bool doBind)
{
    SOCKET sck = socket(AF_INET, SOCK_STREAM, 0);
    if (sck == -1)
    {
        cerr << "Failed to create socket. " << GET_SOCKET_ERROR << endl;
        return -1;
    }
    
    pin->sin_port = htons(my_port);
    pin->sin_addr.s_addr = my_host;
    pin->sin_family = AF_INET;
    
    if (doBind)
    {
        if (bind(sck, &socaddr, sockaddr_len) != 0)
        {
            cerr << "Failed to bind socket. " << GET_SOCKET_ERROR << endl;
            return -1;
        }
    }
    
    return sck;
}

int do_server()
{
    vector<SOCKET> my_events(1);
    vector<struct sockaddr_in> my_address(1);
    my_events[0] = create_socket(true);
    my_address[0] = *pin;
    set_async_mode(my_events[0], true);

    if (listen(my_events[0], 10) != 0)
    {
        cerr << "Failed to listen to server socket! " << GET_SOCKET_ERROR << endl;
        return -1;
    }

    cout << "Ready, waiting for client connection..." << endl;
    
again:

    while(do_continue)
    {   
        struct timeval tv;
        tv.tv_sec = poll_delay;
        tv.tv_usec = 0;

        fd_set rd_set;
        FD_ZERO(&rd_set);
        SOCKET high_sck = 0;
        for(unsigned index = 0; index < my_events.size(); index++)
        {
            FD_SET(my_events[index], &rd_set);
            if (high_sck < my_events[index])
                high_sck = my_events[index];
        }
        ++high_sck;

        int result = select(high_sck, &rd_set, NULL, NULL, &tv);
        if (result > 0)
        {
            vector<SOCKET>::iterator P_event;
            vector<struct sockaddr_in>::iterator P_address;
            unsigned index = 0;
            for(index = 0,
                P_event = my_events.begin(),
                P_address = my_address.begin()
                ; index < my_events.size()
                ; index++,
                ++P_event,
                ++P_address)
            {
                if (FD_ISSET(my_events[index], &rd_set))
                {
                    if (0 == index)
                    {
                        // Listening socket
                        socklen_t sockaddr_len = sizeof(sockaddr);
                        SOCKET conn = accept(my_events[0], &socaddr, &sockaddr_len);
                        if (-1 == conn)
                        {
                            cerr << "Connection failed. " << GET_SOCKET_ERROR << endl;
                            continue;
                        }
                        
                        cout << "Client connected from " 
                            << inet_ntoa(pin->sin_addr) 
                            << ":" 
                            << ntohs(pin->sin_port) 
                            << endl;
                        
                        BOOL opt = true;
                        setsockopt(conn, IPPROTO_TCP, TCP_NODELAY, (const char *)&opt, sizeof(BOOL));
                        set_async_mode(my_events[index], true);

                        my_events.push_back(conn);
                        my_address.push_back(*pin);
                        goto again;
                    }
                    else
                    {
                        // data conn
                        static char buffer[1024];
                        int bytes = 0;
                        if (debug_flag) 
                        {
                            cout << "Receiving from client " 
                                << inet_ntoa(my_address[index].sin_addr) 
                                << ":" 
                                << ntohs(my_address[index].sin_port) 
                                << endl;
                        }

                        while(do_continue)
                        {
                            bytes = recv(my_events[index], buffer, 
                                sizeof(buffer) -1, 0);
                            
                            if (bytes < 0)
                            {
                                if (GET_SOCKET_ERROR != EWOULDBLOCK)
                                {                                    
disconnect:
                                    cout << "Connection closed. " << GET_SOCKET_ERROR << endl;
                                    closesocket(my_events[index]);
                                    
                                    if (log_file.is_open() && log_file)
                                    {
                                        log_file << get_time() << '\t' 
                                            << inet_ntoa(my_address[index].sin_addr) 
                                            << ":"
                                            << (long)ntohs(my_address[index].sin_port) << '\t'
                                            << "disconnected" 
                                            << ENDL;
                                    }

                                    my_events.erase(P_event);
                                    my_address.erase(P_address);
                                }
                                
                                goto again;
                            }
                            else if (bytes == 0)
                                goto disconnect;
                            else
                            {
                                buffer[bytes] = 0;
                                
                                if (debug_flag)
                                {
                                    cout << "Received " << bytes << " bytes. ("
                                        << buffer << ")"
                                        << endl;
                                }
                                if (log_file.is_open() && log_file)
                                {
                                    log_file << get_time() << '\t' 
                                        << inet_ntoa(my_address[index].sin_addr) 
                                        << ":"
                                        << (long)ntohs(my_address[index].sin_port) << '\t'
                                        << "ok" 
                                        << ENDL;
                                }
                                
                                if (strcmp(buffer, "PING\r\n") == 0)
                                {
                                    static const char answer[] = {"PONG\r\n"};
                                    
#ifdef WIN32
                                    if (reply_delay)
                                        Sleep(reply_delay);
#endif
                                    set_async_mode(my_events[index], false);
                                    int bytes = send(my_events[index], answer, strlen(answer), 0);
                                    set_async_mode(my_events[index], true);
                                    
                                    if (debug_flag)
                                    {
                                        cout << "Replyed (pong) with " << bytes << " bytes.";
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    
    return 0;
}


int do_client()
{
    long elapsed_ms = 0;
    
    while(do_continue)
    {
        sleep(loop_delay);
        SOCKET server = create_socket(false);
        
        BOOL opt = true;
        setsockopt(server, IPPROTO_TCP, TCP_NODELAY, (const char *)&opt, sizeof(BOOL));
        
        
        cout << "Connecting to "
            << inet_ntoa(pin->sin_addr) 
            << ":" 
            << ntohs(pin->sin_port) 
            << " ...";
        cout.flush();
        
        
        if (connect(server, &socaddr, sizeof(socaddr)) == -1)
        {
            cout << " FAILED " << GET_SOCKET_ERROR << endl;
            
            if (log_file.is_open())
                log_file << get_time() << '\t' << "no_connect" << ENDL;
            
            continue;
        }
        
        
        cout << " ok" << endl;
        
        if (log_file.is_open())
            log_file << get_time() << '\t' <<"connected" << ENDL;
        
        while(do_continue)
        {
#ifdef WIN32
            SYSTEMTIME st_begin;
            GetSystemTime(&st_begin);
#else
            struct timeval tv_begin;
            gettimeofday(&tv_begin, NULL);
#endif
            
            
            static const char ping[] = {"PING\r\n"};
            const int len = strlen(ping);
            
            if (send(server, ping, len, 0) != len)
            {
                cout << "send() failed. " << GET_SOCKET_ERROR << endl;
                
                if (log_file.is_open())
                    log_file << get_time() << '\t' << "disconnected" << ENDL;
                
                break;
            }
            
            char return_buf[1024];
            if (recv(server, return_buf, 6, 0) < 1)
            {
                cout << "receive() failed. " << GET_SOCKET_ERROR << endl;
                
                if (log_file.is_open())
                    log_file << get_time() << '\t' << "disconnected" << ENDL;
                
                break;
            }
            
#ifdef WIN32
            {
                SYSTEMTIME st_end;
                GetSystemTime(&st_end);
                
                FILETIME ft_begin, ft_end;
                SystemTimeToFileTime(&st_begin, &ft_begin);
                SystemTimeToFileTime(&st_end, &ft_end);
                
                ULARGE_INTEGER li_begin, li_end;
                
                li_begin.LowPart = ft_begin.dwLowDateTime;
                li_begin.HighPart = ft_begin.dwHighDateTime;
                
                li_end.LowPart = ft_end.dwLowDateTime;
                li_end.HighPart = ft_end.dwHighDateTime;
                
                
                elapsed_ms = ((li_end.QuadPart - li_begin.QuadPart) / 10000);
            }
            
#else
            {
                struct timeval tv_end;
                gettimeofday(&tv_end, NULL);
                long long ll_begin = (tv_begin.tv_sec) * 1000 + (tv_begin.tv_usec / 1000);
                long long ll_end = (tv_end.tv_sec) * 1000 + (tv_end.tv_usec / 1000);
                elapsed_ms = ll_end - ll_begin;
            }
#endif
            
            if (verbose_flag)
                cout << "ping/pong in " << elapsed_ms << " milliseconds" << endl;
            
            if (log_file.is_open())
            {
                log_file << get_time() << '\t' 
                    << "ok" << '\t'
                    << elapsed_ms
                    << ENDL;
            }
            
            sleep(loop_delay);
        }
        
        closesocket(server);    
    }
    
    return 0;
}

